#ifndef COMPILER_HEADER
#define COMPILER_HEADER

#include "TypeDefs.h"
#include "Absyn.h"
#include "Skeleton.h"
#include <filesystem>
#include <utility>
#include "codeGen/Variable.h"
#include "ParseError.h"
#include "Info.h"
#include "codeGen/Quadruple.h"
#include "codeGen/BasicBlock.h"
#include "codeGen/QuadrupleGenerator.h"
#include <unordered_set>

using Op = Quadruple::Op;
class Compiler {
public:
	Compiler(const std::string& f, shared_ptr<Scope> s) : file(f), buf(), scope(s), quadGen(std::move(s)) {};

	static const std::string extension;
	static void externalCommand(std::filesystem::path file);

	string compile(Visitable *v);

	static std::string getVirtName(const ClassInfoPtr& c) {
		return c->name + ":virt_table";
	}

	void generateQJumpCond(QJumpCond &q);
	void generateQAlloc(QAlloc &q);
	void generateQWrite(QWrite &q);
	void generateQAccess(QAccess &q);
	void generateQReturn(QReturn &q);
	void generateQCall(QCall &q);
	void generateQParam(QParam &q);
	void generateQJump(QJump &q);
	void generateQAssign(QAssign &q);
	void generateQLabel(QLabel &q);
private:
	std::filesystem::path file;
	std::stringstream buf;
	shared_ptr<Scope> scope;
	QuadrupleGenerator quadGen;

	unordered_set<VariablePtr> temps;
	unordered_set<VarInfoPtr> localInfos;

	void compileFunction(const FunctionInfoPtr& f);
	void printFunction(QuadrupleGenerator::Result quadEnv);

	class Register {
	public:
		Register(string n) : name(std::move(n)) {};
		Register(int n) { name = Compiler::regs[n]; }
		Register(const VariablePtr &v) { name = Compiler::regs[v->registerColor + 1]; }
		string name;
		operator string() { return "%" + name; }
		bool operator==(const Register &other) const { return name == other.name; }
		bool operator!=(const Register &other) const { return name != other.name; }
	};

	static std::string mangleFunctionName(const FunctionInfoPtr& f) {
		if (auto c = f->klass.lock()) {
			return c->name + "::" + f->name;
		}
		return f->name;
	}
	string getRef(VariablePtr &v);


	static const char* const regs[6];

	static string getOpName(Quadruple::Op op) {
		switch(op.op) {
			case Op::Plus: return "ADDL";
			case Op::Minus: return "SUBL";
			case Op::Mul: return "IMULL";
			case Op::Div:
			case Op::Mod: throw runtime_error("Impossible to handle division with default");
			case Op::LT: return "SETL";
			case Op::LE: return "SETLE";
			case Op::EQ: return "SETE";
			case Op::NEQ: return "SETNE";
			case Op::GE: return "SETGE";
			case Op::GT: return "SETG";
			default: return "Op" + to_string(op.op) + "?";
		}
	}

	void moveTo(VariablePtr &v, Register tg) {
		if (v->constExpr) {
			append("MOVL", "$"+to_string(v->val), tg);
			return;
		}
		auto reg = Register(v);
		if (reg == 0) {
			append("MOVL", getRef(v), tg);
		} else if (reg != tg) {
			append("MOVL", reg, tg);
		}
	}

	void moveTo(Register reg, Register tg) {
		if (reg != tg) {
			append("MOVL", reg, tg);
		}
	}

	Register moveToAnyRegister(VariablePtr &v) {
		auto reg = Register(v);
		if (reg == 0) {
			append("MOVL", getRef(v), reg);
		}
		return reg;
	}

	void append(Quadruple::Op op, string op1, string op2 = "", string op3 = "") {
		return append(getOpName(op), op1, op2, op3);
	}

	void append(string cmd, string op1 = "", string op2 = "", string op3 = "") {
		buf << "\t" << cmd << " " << op1 << " " << op2 << " " << op3 << endl;
	}
};

#endif
