#include "RegisterAllocator.h"
#include "BasicBlock.h"
#include "setOverloads.h"
#include <fstream>

void RegisterAllocator::analyseLive() {
	for (const auto& v : vars) {
		if (!v->info) continue;
		for (const auto& q : v->writes) {
			q->block->flow.addDef(v);
		}
		for (const auto& q : v->uses) {
			q->block->flow.addUse(v);
		}
	}

	for (const auto& b : blocks) {
		for (const auto& phiMap : b->phi->phi) {
			for (const auto& phiRow : phiMap) {
				b->flow.def.emplace(phiRow.first);
			}
		}
	}

	while (true) {
		bool changed = false;
		for (const auto& b : blocks) {
			set<VariablePtr> pIn, pOut, diff;
			auto &f = b->flow;
			pIn.swap(f.in);
			pOut.swap(f.out);
			f.in = f.use + (pOut - f.def);
			for (const auto& succ : b->out) {
				f.out += succ->flow.inForBlock(b);
			}
			if (f.in != pIn || f.out != pOut) changed = true;
		}
		if (!changed) break;
	}

	if (!blocks.empty() && !blocks[0]->flow.in.empty())
		throw ParseError("Variable definition not found - first block expects input");
}

void RegisterAllocator::allocate() {
	analyseLive();
	buildGraph();
}

void printGraph(const Graph<VariablePtr> &graph, const std::string& filename) {
	ofstream f(filename);
	for (const auto &v : graph.getEdges()) {
		f << v.first->name << " " << v.second->name << endl;
	}
	f.close();
}

void printGraph(const Graph<std::string> &graph, const std::string& filename) {
	ofstream f(filename);
	for (const auto &v : graph.getEdges()) {
		f << v.first << " " << v.second << endl;
	}
	f.close();
}

void RegisterAllocator::buildGraph() {
	bool anyAlive = false;
	for (const auto& b : blocks) {
		set<VariablePtr> alive = b->flow.in;
		for (const auto& q : b->quads) {
			if (auto assign = dynamic_pointer_cast<QAssign>(q)) {
				if (assign->op == Quadruple::Op::Copy && !assign->args[0]->constExpr) {
					graph.addAssign(assign->loc, assign->args[0]);
					g2.addAssign(assign->loc->name, assign->args[0]->name);
				}
			}
			alive += to_set(q->definitions());
			vector<VariablePtr> dead;
			for (const auto& var : alive) {
				// this quad is the last quad that uses var
				if ((var->uses.empty() || var->uses.back() == q) && !b->flow.out.contains(var)) {
					dead.emplace_back(var);
				}
			}
			alive -= to_set(dead);
			q->aliveAfter = alive;
			anyAlive |= alive.size() >= 1;
			for (auto i : alive) {
				for (auto j : alive) {
					if (i != j) {
						graph.add(i, j);
						g2.add(i->name, j->name);
					}
				}
			}
		}
	}
	if (!anyAlive) return;
	graph.eval(5);
	g2.eval(5);
	int gNumSpill = 0, g2NumSpill = 0;
	for (auto &v : graph.getAllocation()) {
		v.first->registerColor = v.second;
		if (v.second == -1) {
			gNumSpill++;
		} else {
			v.first->name += "-r" + to_string(v.second);
		}
	}
	for (auto &v : g2.getAllocation()) {
		if (v.second == -1)
			g2NumSpill++;
	}
	cout << "allocation: VPtr eval spills: " << gNumSpill << " vs strings spills: " << g2NumSpill << endl;
	if (!g2.isChordal()) {
		printGraph(graph, "./mygraph");
		printGraph(g2, "./myg2");
		if (graph.isChordal()) {
			throw ParseError("allocation: g2 not chordal, tho g1 is chordal!");
		}
		throw ParseError("allocation: none graph chordal!");
	} else {
		if (!graph.isChordal()) {
			throw ParseError("allocation: g1 not chordal, tho g2 is chordal!");
		}
	}
}


